@@ -6,15 +6,33 @@
 # -- Project information -----------------------------------------------------
 # https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
 
-import os
+from pathlib import Path
 import sys
 
-sys.path.insert(0, os.path.abspath('../..'))
-from libpulse import __version__
+__version__ = 'Unknown version'
+def conf_py_setup():
+    global __version__
+
+    cur_dir = Path(__file__).parent
+    root_dir = cur_dir.parent.parent
+    try:
+        sys.path.append(str(root_dir))
+        from libpulse import __version__
+    finally:
+        sys.path.pop()
+
+    with open(cur_dir / 'README.rst', 'w') as fdest:
+        fdest.write('libpulse |version|\n')
+        fdest.write('==================\n\n')
+        with open(root_dir / 'README.rst') as fsrc:
+            content = fsrc.read()
+            fdest.write(content)
+
+conf_py_setup()
 
 project = 'libpulse'
-copyright = '2024, Xavier de Gaye'
-author = 'Xavier de Gaye'
+copyright = '2025, Xavier de Gaye'
+author = ''
 
 # The short X.Y version
 version = __version__
@@ -32,7 +50,7 @@ exclude_patterns = []
 # https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
 
 html_theme = 'sphinx_rtd_theme'
-html_static_path = ['_static']
+html_static_path = ['images']
 
 # -- Options for manual page output ------------------------------------------
 
