@@ -6,9 +6,11 @@ import logging
 import re
 import pprint
 import ctypes as ct
+from abc import ABC
 from functools import partialmethod
 
-from .libpulse_ctypes import PulseCTypes, PA_INVALID_INDEX
+from . import __version__
+from .libpulse_ctypes import PulseCTypes
 from .mainloop import MainLoop, pulse_ctypes, callback_func_ptr
 from .pulse_functions import pulse_functions
 from .pulse_enums import pulse_enums
@@ -34,15 +36,32 @@ def _add_pulse_to_namespace():
 _add_pulse_to_namespace()
 del _add_pulse_to_namespace
 
+# /usr/include/pulse/def.h:
+# #define PA_INVALID_INDEX ((uint32_t) -1)
+PA_INVALID_INDEX = ct.c_uint32(-1).value
+
+# 'pa_volume_t' values defined in volume.h.
+PA_VOLUME_NORM =    0x10000    # Normal volume (100%, 0 dB).
+PA_VOLUME_MUTED =   0          # Muted (minimal valid) volume (0%, -inf dB).
+PA_VOLUME_MAX =     0xffffffff # UINT32_MAX/2 Maximum volume we can store.
+PA_VOLUME_INVALID = 0xffffffff # Special 'invalid' volume.
+def PA_VOLUME_IS_VALID(v): return v <= PA_VOLUME_MAX
+
+PA_CHANNELS_MAX = 32           # Defined in sample.h
+C_UINT_ARRAY_32 = ct.c_uint * PA_CHANNELS_MAX
+C_INT_ARRAY_32 = ct.c_int * PA_CHANNELS_MAX
+
 # Map values to their name.
-ERROR_CODES = dict((eval(var), var) for var in globals() if
-                   var.startswith('PA_ERR_') or var == 'PA_OK')
 CTX_STATES = dict((eval(state), state) for state in
                   ('PA_CONTEXT_UNCONNECTED', 'PA_CONTEXT_CONNECTING',
                    'PA_CONTEXT_AUTHORIZING', 'PA_CONTEXT_SETTING_NAME',
                    'PA_CONTEXT_READY', 'PA_CONTEXT_FAILED',
                    'PA_CONTEXT_TERMINATED'))
 
+OPERATION_STATES = dict((eval(state), state) for state in
+                        ('PA_OPERATION_CANCELLED', 'PA_OPERATION_DONE',
+                         'PA_OPERATION_RUNNING'))
+
 def event_codes_to_names():
     def build_events_dict(mask):
         for fac in globals():
@@ -66,7 +85,23 @@ def run_in_task(coro):
     """Decorator to wrap a coroutine in a task of AsyncioTasks instance."""
 
     async def wrapper(*args, **kwargs):
-        lib_pulse = LibPulse.get_instance()
+        def get_coro_arg():
+            length = len(args)
+            coro_arg = ''
+            if length >=2:
+                coro_arg += f'{args[1].__qualname__}('
+                if length >= 3:
+                    coro_arg += f'{args[2]})'
+                else:
+                    coro_arg += ')'
+            return coro_arg
+
+        if 0:
+            # When enabled while running the test suite, will print all the
+            # pulseaudio coroutines that are being tested.
+            print(f'{coro.__qualname__}({get_coro_arg()})', file=sys.stderr)
+
+        lib_pulse = LibPulse._get_instance()
         if lib_pulse is None:
             raise LibPulseClosedError
 
@@ -74,7 +109,8 @@ def run_in_task(coro):
             return await lib_pulse.libpulse_tasks.create_task(
                                                     coro(*args, **kwargs))
         except asyncio.CancelledError:
-            logger.error(f'{coro.__qualname__}() has been cancelled')
+            logger.warning(f'{coro.__qualname__}({get_coro_arg()})'
+                                                    ' has been cancelled')
             raise
     return wrapper
 
@@ -188,12 +224,12 @@ class PropList(dict):
         null_ptr_ptr = ct.pointer(null_ptr)
         while True:
             key = pa_proplist_iterate(c_pa_proplist, null_ptr_ptr)
-            if isinstance(key, bytes):
-                val = pa_proplist_gets(c_pa_proplist, key)
-                if bool(val):
-                    self[key.decode()] = val.decode()
-            elif not bool(key):
+            if not key:
                 break
+            elif isinstance(key, bytes):
+                val = pa_proplist_gets(c_pa_proplist, key)
+                if val:
+                    self[key.decode()] = val.decode()
 
 class PulseStructure:
     """The representation of a ctypes Structure.
@@ -230,7 +266,7 @@ class PulseStructure:
                 setattr(self, name, c_struct_val)
 
             # A NULL pointer.
-            elif not bool(c_struct_val):
+            elif not c_struct_val:
                 setattr(self, name, None)
 
             elif c_type is ct.c_char_p:
@@ -249,20 +285,16 @@ class PulseStructure:
             # An array of pointers.
             elif (isinstance(c_struct_val, ct._Pointer) and
                     isinstance(c_struct_val.contents, ct._Pointer)):
-                ptr = c_struct_val.contents
-                ctype = ptr._type_
+                ctype = c_struct_val.contents._type_
                 if ctype.__name__ in struct_ctypes:
                     val = []
+                    ptr = c_struct_val
                     size_attr = self.array_sizes[ctype.__name__]
                     array_size = getattr(self, size_attr)
-                    i = 0
-                    while True:
-                        if not bool(ptr[i]):
+                    for i in range(array_size):
+                        if not ptr[i]:
                             break
-                        if i == array_size:
-                            break
-                        val.append(PulseStructure(ptr[i], ctype))
-                        i += 1
+                        val.append(PulseStructure(ptr[i].contents, ctype))
                     setattr(self, name, val)
                 else:
                     self.ignore_member(fq_name)
@@ -291,6 +323,89 @@ class PulseStructure:
     def __repr__(self):
         return pprint.pformat(self.__dict__, sort_dicts=False)
 
+class CtypesPulseStructure(ABC):
+    """Container for an instance of a subclass of ctypes.Structure.
+
+    Scalar attributes of the instance may be updated after instanciation as
+    well as the scalar elements of an array, but not the array itself or the
+    other aggregate attributes.
+    """
+
+    @staticmethod
+    def get_fields_names(struct_name):
+        return [field[0] for field in struct_ctypes[struct_name]._fields_]
+
+    def __getattr__(self, name):
+        if name in self.fields_names:
+            return getattr(self.ct_struct, name)
+        else:
+            raise AttributeError(
+              f"'{self.__class__.__name__}' object has no attribute '{name}'")
+
+    def __setattr__(self, name, val):
+        if name in self.fields_names:
+            setattr(self.c_obj, name, val)
+        else:
+            object.__setattr__(self, name, val)
+
+    def byref(self):
+        return ct.byref(self.ct_struct)
+
+    def to_pulse_structure(self):
+        return PulseStructure(self.ct_struct, struct_ctypes[self.struct_name])
+
+class Pa_buffer_attr(CtypesPulseStructure):
+
+    struct_name = 'pa_buffer_attr'
+    fields_names = CtypesPulseStructure.get_fields_names(struct_name)
+
+    def __init__(self, maxlength, tlength, prebuf, minreq, fragsize):
+        self.ct_struct = struct_ctypes[self.struct_name](
+                                maxlength, tlength, prebuf, minreq, fragsize)
+
+class Pa_cvolume(CtypesPulseStructure):
+
+    struct_name = 'pa_cvolume'
+    fields_names = CtypesPulseStructure.get_fields_names(struct_name)
+
+    def __init__(self, channels, values):
+        length = len(values)
+        assert length <= PA_CHANNELS_MAX
+        values += [0 for i in range(PA_CHANNELS_MAX - length)]
+        self.ct_struct = struct_ctypes[self.struct_name](
+                                        channels, C_UINT_ARRAY_32(*values))
+
+class Pa_channel_map(CtypesPulseStructure):
+
+    struct_name = 'pa_channel_map'
+    fields_names = CtypesPulseStructure.get_fields_names(struct_name)
+
+    def __init__(self, channels, map):
+        length = len(map)
+        assert length <= PA_CHANNELS_MAX
+        map += [0 for i in range(PA_CHANNELS_MAX - length)]
+        self.ct_struct = struct_ctypes[self.struct_name](
+                                            channels, C_INT_ARRAY_32(*map))
+
+class Pa_format_info(CtypesPulseStructure):
+
+    struct_name = 'pa_format_info'
+    fields_names = CtypesPulseStructure.get_fields_names(struct_name)
+
+    def __init__(self, encoding, plist):
+        # 'plist' is an instance of ctypes._Pointer that has been built using
+        # some of the pa_proplist_ functions.
+        self.ct_struct = struct_ctypes[self.struct_name](encoding, plist)
+
+class Pa_sample_spec(CtypesPulseStructure):
+
+    struct_name = 'pa_sample_spec'
+    fields_names = CtypesPulseStructure.get_fields_names(struct_name)
+
+    def __init__(self, format, rate, channels):
+        self.ct_struct = struct_ctypes[self.struct_name](
+                                                    format, rate, channels)
+
 class LibPulse:
     """Interface to libpulse library as an asynchronous context manager."""
 
@@ -300,12 +415,26 @@ class LibPulse:
     #                      [pa_context *, [args...], cb_t, void *])
     # Callback signature: (void, [pa_context *, [objs...], void *])
     context_methods = (
+        'pa_context_add_autoload',
         'pa_context_drain',
         'pa_context_get_server_info',
         'pa_context_load_module',
         'pa_context_play_sample_with_proplist',
+
+        # pa_context_send_message_to_object() is the only method that is not
+        # in the 'LibPulse.context_list_methods' list and that returns a list
+        # of objects. See the 'pa_context_string_cb_t' callback signature in
+        # the pulse_functions module.
         'pa_context_send_message_to_object',
+
+        # The context state is monitored by the LibPulse instance.
+        # 'pa_context_set_state_callback',
+
+        # Reception of events is monitored by the LibPulse instance.
+        # 'pa_context_set_subscribe_callback',
+
         'pa_context_stat',
+        'pa_ext_device_restore_test',
     )
 
     # Function signature: (pa_operation *,
@@ -323,6 +452,8 @@ class LibPulse:
         'pa_context_play_sample',
         'pa_context_proplist_remove',
         'pa_context_proplist_update',
+        'pa_context_remove_autoload_by_index',
+        'pa_context_remove_autoload_by_name',
         'pa_context_remove_sample',
         'pa_context_set_card_profile_by_index',
         'pa_context_set_card_profile_by_name',
@@ -352,11 +483,16 @@ class LibPulse:
         'pa_context_suspend_source_by_index',
         'pa_context_suspend_source_by_name',
         'pa_context_unload_module',
+        'pa_ext_device_restore_save_formats',
+        'pa_ext_device_restore_subscribe',
     )
 
     # Function signature: (pa_operation *, [pa_context *, cb_t, void *])
     # Callback signature: (void, [pa_context *, struct *, int, void *])
     context_list_methods = (
+        'pa_context_get_autoload_info_by_index',
+        'pa_context_get_autoload_info_by_name',
+        'pa_context_get_autoload_info_list',
         'pa_context_get_card_info_by_index',
         'pa_context_get_card_info_by_name',
         'pa_context_get_card_info_list',
@@ -377,6 +513,8 @@ class LibPulse:
         'pa_context_get_source_info_list',
         'pa_context_get_source_output_info',
         'pa_context_get_source_output_info_list',
+        'pa_ext_device_restore_read_formats',
+        'pa_ext_device_restore_read_formats_all',
     )
 
     # Function signature: (pa_operation *,
@@ -396,10 +534,24 @@ class LibPulse:
         'pa_stream_update_timing_info',
     )
 
-    def __init__(self, name):
-        """'name' is the name of the application."""
+    def __init__(self, name, server=None, flags=PA_CONTEXT_NOAUTOSPAWN):
+        """ Constructor arguments:
 
+            - 'name'    Name of the application.
+            - 'server'  Server name, if 'server' is None, connect to the
+                        default server.
+            - 'flags'   'flags' and 'server' are arguments of
+                        pa_context_connect() used to connect to the server.
+        """
+
+        logger.info(f'Python libpulse version {__version__}')
         assert isinstance(name, str)
+        if server is not None:
+            assert isinstance(server, str)
+            server = server.encode()
+        self.server = server
+        self.flags = flags
+
         self.loop = asyncio.get_running_loop()
         if self.loop in self.ASYNCIO_LOOPS:
             raise LibPulseInstanceExistsError
@@ -408,7 +560,7 @@ class LibPulse:
                                         name.encode())
         # From the ctypes documentation: "NULL pointers have a False
         # boolean value".
-        if not bool(self.c_context):
+        if not self.c_context:
             raise RuntimeError('Cannot get context from libpulse library')
 
         self.closed = False
@@ -447,7 +599,13 @@ class LibPulse:
                     delattr(this_module, func_name)
 
     @staticmethod
-    def get_instance():
+    def _get_instance():
+        """Get the LibPulse instance running on the asyncio loop.
+
+        The instance may not be yet connected or be in the closing state.
+        Prefer using get_current_instance().
+        """
+
         loop = asyncio.get_running_loop()
         try:
             return LibPulse.ASYNCIO_LOOPS[loop]
@@ -458,7 +616,7 @@ class LibPulse:
     def context_state_callback(c_context, c_userdata):
         """Call back that monitors the connection state."""
 
-        lib_pulse = LibPulse.get_instance()
+        lib_pulse = LibPulse._get_instance()
         if lib_pulse is None:
             return
 
@@ -466,13 +624,12 @@ class LibPulse:
         st = CTX_STATES[st]
         if st in ('PA_CONTEXT_READY', 'PA_CONTEXT_FAILED',
                      'PA_CONTEXT_TERMINATED'):
-            error = pa_context_errno(c_context)
-            error = ERROR_CODES[error]
-
-            state = (st, error)
+            error = pa_strerror(pa_context_errno(c_context))
+            state = (st, error.decode())
             logger.info(f'LibPulse connection: {state}')
 
             state_notification = lib_pulse.state_notification
+            lib_pulse.state = state
             if not state_notification.done():
                 state_notification.set_result(state)
             elif not lib_pulse.closed and st != 'PA_CONTEXT_READY':
@@ -489,11 +646,9 @@ class LibPulse:
 
         pa_context_set_state_callback(self.c_context,
                                       self.c_context_state_callback, None)
-        rc = pa_context_connect(self.c_context, None, PA_CONTEXT_NOAUTOSPAWN,
-                                None)
+        rc = pa_context_connect(self.c_context, self.server, self.flags, None)
         logger.debug(f'pa_context_connect return code: {rc}')
         await self.state_notification
-        self.state = self.state_notification.result()
 
         if self.state[0] != 'PA_CONTEXT_READY':
             raise LibPulseStateError(self.state)
@@ -502,7 +657,7 @@ class LibPulse:
     def context_subscribe_callback(c_context, event_type, index, c_userdata):
         """Call back to handle pulseaudio events."""
 
-        lib_pulse = LibPulse.get_instance()
+        lib_pulse = LibPulse._get_instance()
         if lib_pulse is None:
             return
 
@@ -546,12 +701,17 @@ class LibPulse:
         return c_operation
 
     @staticmethod
-    async def handle_operation(c_operation, future, errmsg):
+    async def handle_operation(c_operation, future):
         # From the ctypes documentation: "NULL pointers have a False
         # boolean value".
-        if not bool(c_operation):
+        if not c_operation:
             future.cancel()
-            raise LibPulseOperationError(errmsg)
+            error = "NULL 'pa_operation' pointer"
+            lib_pulse = LibPulse._get_instance()
+            if lib_pulse is not None:
+                errmsg = pa_strerror(pa_context_errno(lib_pulse.c_context))
+                error += f': {errmsg.decode()}'
+            raise LibPulseOperationError(error)
 
         try:
             await future
@@ -571,45 +731,51 @@ class LibPulse:
 
         def callback_func(c_operation_type, *c_results):
             results = []
-            for arg, c_result in zip(callback_sig[1][1:-1], c_results[:-1]):
-                arg_list = arg.split()
-                if arg_list[-1] == '*':
-                    assert arg_list[0] in struct_ctypes
-                    struct_name = arg_list[0]
-                    if not bool(c_result):
-                        results.append(None)
-                    else:
-                        results.append(PulseStructure(c_result.contents,
+            try:
+                for arg, c_result in zip(callback_sig[1][1:-1],
+                                                            c_results[:-1]):
+                    arg_list = arg.split()
+                    if arg_list[-1] == '*' and arg_list[0] in struct_ctypes:
+                        struct_name = arg_list[0]
+                        if not c_result:
+                            results.append(None)
+                        else:
+                            results.append(PulseStructure(c_result.contents,
                                                 struct_ctypes[struct_name]))
-                else:
-                    results.append(c_result)
-
-            if not notification.done():
-                notification.set_result(results)
+                    else:
+                        if arg == 'char *':
+                            c_result = c_result.decode()
+                        results.append(c_result)
+            except Exception as e:
+                results = e
+            finally:
+                if not notification.done():
+                    notification.set_result(results)
 
         callback_data = self.get_callback_data(func_name)
         assert callback_data, f'{func_name} signature without a callback'
         callback_name, callback_sig = callback_data
 
         notification  = self.loop.create_future()
-        errmsg = f'Error at {func_name}()'
 
         # Await on the future.
         cb_func_ptr = callback_func_ptr(callback_name, callback_func)
         c_operation = self.call_ctypes_func(func_name, operation_type,
                                                     cb_func_ptr, *func_args)
-        await LibPulse.handle_operation(c_operation, notification, errmsg)
+        await LibPulse.handle_operation(c_operation, notification)
 
         results = notification.result()
+        if isinstance(results, Exception):
+            raise results
         for result in results:
             if result is None:
-                raise LibPulseOperationError(errmsg)
+                raise LibPulseOperationError(
+                            'NULL pointer result returned by the callback')
         if len(results) == 1:
             return results[0]
         return results
 
-    @run_in_task
-    async def _pa_context_get_list(self, func_name, *func_args):
+    async def _pa_get_list(self, func_name, operation_type, *func_args):
         """Call an asynchronous pulse function that returns a list.
 
         'func_args' is the sequence of the arguments of the function preceding
@@ -617,20 +783,24 @@ class LibPulse:
         (i.e. 'userdata') is set to None by call_ctypes_func().
         """
 
-        def info_callback(c_context, c_info, eol, c_userdata):
+        def info_callback(c_operation_type, c_info, eol, c_userdata):
             # From the ctypes documentation: "NULL pointers have a False
             # boolean value".
-            if not bool(c_info):
+            if not c_info:
                 if not notification.done():
                     notification.set_result(eol)
             else:
-                arg = callback_sig[1][1]
-                arg_list = arg.split()
-                assert arg_list[-1] == '*'
-                assert arg_list[0] in struct_ctypes
-                struct_name = arg_list[0]
-                infos.append(PulseStructure(c_info.contents,
-                                            struct_ctypes[struct_name]))
+                try:
+                    arg = callback_sig[1][1]
+                    arg_list = arg.split()
+                    assert arg_list[-1] == '*'
+                    assert arg_list[0] in struct_ctypes
+                    struct_name = arg_list[0]
+                    infos.append(PulseStructure(c_info.contents,
+                                                struct_ctypes[struct_name]))
+                except Exception as e:
+                    if not notification.done():
+                        notification.set_result(e)
 
         callback_data = self.get_callback_data(func_name)
         assert callback_data, f'{func_name} signature without a callback'
@@ -638,39 +808,67 @@ class LibPulse:
 
         infos = []
         notification  = self.loop.create_future()
-        errmsg = f'Error at {func_name}()'
 
         # Await on the future.
         cb_func_ptr = callback_func_ptr(callback_name, info_callback)
-        c_operation = self.call_ctypes_func(func_name, self.c_context,
+        c_operation = self.call_ctypes_func(func_name, operation_type,
                                                     cb_func_ptr, *func_args)
-        await LibPulse.handle_operation(c_operation, notification, errmsg)
+        await LibPulse.handle_operation(c_operation, notification)
 
         eol = notification.result()
+        if isinstance(eol, Exception):
+            raise eol
         if eol < 0:
-            raise LibPulseOperationError(errmsg)
-        if func_name.endswith(('_by_name', '_by_index')):
+            error = "'eol' set to a negative value by the callback"
+            errmsg = pa_strerror(pa_context_errno(self.c_context))
+            error += f': {errmsg.decode()}'
+            raise LibPulseOperationError(error)
+
+        if func_name.endswith(('_by_name', '_by_index', '_info', '_formats')):
             assert len(infos) == 1
             return infos[0]
         return infos
 
     @run_in_task
+    async def _run_in_task(self, method,
+                           func_name, operation_type, *func_args):
+        try:
+            return await method(func_name, operation_type, *func_args)
+        except (Exception, asyncio.CancelledError) as e:
+            return e
+
     async def _pa_context_get(self, func_name, *func_args):
-        return await self._pa_get(func_name, self.c_context, *func_args)
+        result = await self._run_in_task(self._pa_get, func_name,
+                                         self.c_context, *func_args)
+        if isinstance(result, (Exception, asyncio.CancelledError)):
+            raise result
+        return result
+
+    async def _pa_success(self, func_name, operation_type, *func_args):
+        success = await self._run_in_task(self._pa_get, func_name,
+                                          operation_type, *func_args)
+        if isinstance(success, (Exception, asyncio.CancelledError)):
+            raise success
+        return success
 
-    @run_in_task
     async def _pa_context_op_success(self, func_name, *func_args):
-        success = await self._pa_get(func_name, self.c_context, *func_args)
-        if success == PA_OPERATION_CANCELLED:
-            logger.debug(f'Got PA_OPERATION_CANCELLED for {func_name}')
-        return success
+        result = await self._pa_success(func_name, self.c_context, *func_args)
+        if result != PA_OPERATION_DONE:
+            error = pa_strerror(pa_context_errno(self.c_context))
+            raise LibPulseOperationError(f'Failure: {error.decode()}: '
+                                         f'{OPERATION_STATES[result]!r}')
+        return result
 
-    @run_in_task
-    async def _pa_stream_op_success(self, func_name, stream, *func_args):
-        success = await self._pa_get(func_name, stream, *func_args)
-        if success == PA_OPERATION_CANCELLED:
-            logger.debug(f'Got PA_OPERATION_CANCELLED for {func_name}')
-        return success
+    async def _pa_stream_op_success(self, func_name, pa_stream, *func_args):
+        return await self._pa_success(func_name, pa_stream, *func_args)
+
+    async def _pa_context_get_list(self, func_name, *func_args):
+        result = await self._run_in_task(self._pa_get_list, func_name,
+                                         self.c_context, *func_args)
+        if isinstance(result, (Exception, asyncio.CancelledError)):
+            raise result
+        else:
+            return result
 
 
     # Context manager.
@@ -679,7 +877,7 @@ class LibPulse:
         logger.error(f'The LibPulse instance has been aborted: {state}')
         self.main_task.cancel()
 
-    def close(self):
+    async def close(self):
         if self.closed:
             return
         self.closed = True
@@ -694,6 +892,12 @@ class LibPulse:
             pa_context_set_subscribe_callback(self.c_context, None, None)
 
             if self.state[0] == 'PA_CONTEXT_READY':
+                try:
+                    await self.pa_context_drain()
+                except (Exception, asyncio.CancelledError):
+                    # Quoting pulse documentation:
+                    # If there is nothing to drain, the function returns NULL.
+                    pass
                 pa_context_disconnect(self.c_context)
                 logger.info('Disconnected from libpulse context')
         finally:
@@ -719,22 +923,37 @@ class LibPulse:
                                     self.c_context_subscribe_callback, None)
             return self
         except asyncio.CancelledError:
-            self.close()
+            await self.close()
             if self.state[0] != 'PA_CONTEXT_READY':
                 raise LibPulseStateError(self.state)
         except Exception:
-            self.close()
+            await self.close()
             raise
 
     async def __aexit__(self, exc_type, exc_value, traceback):
-        self.close()
+        await self.close()
         if exc_type is asyncio.CancelledError:
             if self.state[0] != 'PA_CONTEXT_READY':
                 raise LibPulseStateError(self.state)
 
 
     # Public methods.
-    def get_events(self):
+    @staticmethod
+    async def get_current_instance():
+        """Get the LibPulse running instance.
+
+        Raises LibPulseStateError if the instance is not in the
+        PA_CONTEXT_READY state.
+        """
+
+        lib_pulse = LibPulse._get_instance()
+        if lib_pulse is not None:
+            await lib_pulse.state_notification
+            if lib_pulse.state[0] != 'PA_CONTEXT_READY':
+                raise LibPulseStateError(lib_pulse.state)
+        return lib_pulse
+
+    def get_events_iterator(self):
         """Return an Asynchronous Iterator of libpulse events.
 
         The iterator is used to run an async for loop over the PulseEvent
