@@ -5,7 +5,8 @@ from unittest import TestCase, mock
 from contextlib import redirect_stdout
 
 # Load the tests in the order they are declared.
-from . import requires_resources, load_ordered_tests
+from . import load_ordered_tests as load_tests
+from . import requires_resources
 
 import libpulse.libpulse_ctypes as libpulse_ctypes_module
 from ..libpulse_ctypes import PulseCTypes, PulseCTypesLibError
